#include <stdio.h>
#include <string.h>

#include "gg.h"

int apply(FILE *fptr, int prg_len, int mapper, unsigned short address, unsigned char val, int compare);

int main(int argc, char **argv)
{
	char *filename, *code;
	unsigned char magic[4];
	unsigned char num_prg_banks, num_chr_banks, mapper, mapper1, reserved[8];
	FILE *fptr;
	long prg_offset;
	int i;

	if (argc<3) {
		printf("Usage: ./patchnes rom.nes code1 ... [code n]\n");
		return 1;
	}

	filename = argv[1];
	printf("Working on rom %s\n", filename);
	
	fptr = fopen(filename, "r+b");
	if (!fptr) { perror("fopen"); return 1; }

	fread(magic, 4, 1, fptr);
	if (!(magic[0] == 'N' &&
			magic[1] == 'E' &&
			magic[2] == 'S' &&
			magic[3] == 0x1a))
	{
		fprintf(stderr, "Bad header\n");
		return 1;
	}

	fread(&num_prg_banks, 1, 1, fptr);
	fread(&num_chr_banks, 1, 1, fptr);
	fread(&mapper, 1, 1, fptr);
	fread(&mapper1, 1, 1, fptr);
	fread(reserved, 8, 1, fptr);

	prg_offset = ftell(fptr);
	
	switch(mapper>>4)
	{
		case 0:
		case 1:
			break;
		default:
			fprintf(stderr, "Mapper type %d not supported\n", mapper>>4);
			fclose(fptr);
			return 1;
	}

	for (i=2; i<argc; i++)
	{
		unsigned short address;
		unsigned char val, cmp;
		code = argv[i];
		switch (strlen(code))
		{
			case 8:
				printf("Applying 8 letters code '%s'\n", code);
				gg_decode8letters(code, &address, &val, &cmp);
				fseek(fptr, prg_offset, SEEK_SET);
				apply(fptr, num_prg_banks*16*1024, mapper>>4, address, val, cmp);
				break;
			case 6:
				printf("Applying 6 letters code '%s'\n", code);
				gg_decode6letters(code, &address, &val);
				fseek(fptr, prg_offset, SEEK_SET);
				apply(fptr, num_prg_banks*16*1024, mapper>>4, address, val, -1);
				break;
			default:
			printf("Skipping invalid code '%s'\n", code);
				
		}
		
	}
	
	
	
	fclose(fptr);

	return 0;	
}

// call with fptr pointed at prg start
int apply(FILE *fptr, int prg_len, int mapper, unsigned short address, unsigned char val, int compare)
{
	unsigned char tmp;
	long begin_offset;
	int i;
	
	begin_offset = ftell(fptr);
	
	switch(mapper)
	{
		case 0:
			fseek(fptr, address>>1, SEEK_CUR);
				fread(&tmp, 1, 1, fptr);
			if (compare>=0 && tmp != compare) {
				printf("Not replacing $%02X by %02X at address $%04X because it is not $%02X\n", tmp, val, address, compare);
			}
			else
			{
				if (tmp == compare) {
					fseek(fptr, -1, SEEK_CUR);
					fwrite(&val, 1, 1, fptr);
					printf("Replaced $%02X by $%02X at address $%04X\n", tmp, val, address);
				}
			}
			break;
		case 1:
			for (i=0; i<prg_len; i+=0x4000)
			{
				fseek(fptr, begin_offset + i, SEEK_SET);

				// dont overflow on chr rom
				if (begin_offset + i + 0x4000 > begin_offset + prg_len) { break; }
			fseek(fptr, address>>1, SEEK_CUR);
				fread(&tmp, 1, 1, fptr);
			if (compare>=0 && tmp != compare) {
				printf("Not replacing $%02X by %02X at address $%04X because it is not $%02X\n", tmp, val, address + i, compare);
			}
			else
			{
				if (tmp == compare) {
					fseek(fptr, -1, SEEK_CUR);
					fwrite(&val, 1, 1, fptr);
					printf("Replaced $%02X by $%02X at address $%04X\n", tmp, val, address + i);
				}
			}

			}
			break;
	}
	
	return 0;
}


